# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

from mock import patch

from pyJoules.energy_meter import EnergyMeter, EnergySample
from pyJoules.energy_meter import EnergyMeterNotStartedError, EnergyMeterNotStoppedError, SampleNotFoundError
from pyJoules.energy_device import EnergyDevice, EnergyDomain
from ...utils.sample import assert_sample_are_equals

DEVICE1_ENERGY_TRACE = [[1.0, 1.1],
                        [2.0, 2.2],
                        [3.0, 3.4],
                        [4.0, 4.6],
                        [5.0, 5.8]]

DEVICE2_ENERGY_TRACE = [[4.0],
                        [6.2],
                        [8.0],
                        [10.3],
                        [11.123]]

TIMESTAMP_TRACE = [1.1, 2.2, 3.3, 4.4, 5.5]


class EnergyDomainDevice1Domain1(EnergyDomain):
    def __repr__(self):
        return 'device1_domain1'


class EnergyDomainDevice1Domain2(EnergyDomain):
    def __repr__(self):
        return 'device1_domain2'


class EnergyDomainDevice2Domain1(EnergyDomain):
    def __repr__(self):
        return 'device2_domain1'


class MockedEnergyDevice1(EnergyDevice):

    def __init__(self):
        EnergyDevice.__init__(self)
        self.iterator = DEVICE1_ENERGY_TRACE.__iter__()

    @staticmethod
    def available_domains():
        return [EnergyDomainDevice1Domain1(), EnergyDomainDevice1Domain2()]

    def configure(self, domains=None):
        EnergyDevice.configure(self, domains)

    def get_energy(self):
        return self.iterator.__next__()


class MockedEnergyDevice2(EnergyDevice):

    def __init__(self):
        EnergyDevice.__init__(self)
        self.iterator = DEVICE2_ENERGY_TRACE.__iter__()

    @staticmethod
    def available_domains():
        return [EnergyDomainDevice2Domain1()]

    def configure(self, domains=None):
        EnergyDevice.configure(self, domains)

    def get_energy(self):
        return self.iterator.__next__()


@pytest.fixture
def energy_meter():
    device1 = MockedEnergyDevice1()
    device1.configure()
    device2 = MockedEnergyDevice2()
    device2.configure()
    return EnergyMeter([device1, device2])


############################
# NON STARTED ERRORS TESTS #
############################
def test_record_on_non_started_energy_meter_raise_EnergyMeterNotStartedError(energy_meter):
    with pytest.raises(EnergyMeterNotStartedError):
        energy_meter.record()


def test_stop_a_non_started_energy_meter_raise_EnergyMeterNotStartedError(energy_meter):
    with pytest.raises(EnergyMeterNotStartedError):
        energy_meter.stop()


def test_iter_on_a_non_started_energy_meter_raise_EnergyMeterNotStartedError(energy_meter):
    with pytest.raises(EnergyMeterNotStartedError):
        energy_meter.__iter__()


def test_get_sample_on_a_non_started_energy_meter_raise_EnergyMeterNotStartedError(energy_meter):
    with pytest.raises(EnergyMeterNotStartedError):
        energy_meter.get_sample('toto')


############################
# NON STOPPED ERRORS TESTS #
############################
def test_iter_on_a_non_stopped_energy_meter_raise_EnergyMeterNotStoppedError(energy_meter):
    energy_meter.start()
    with pytest.raises(EnergyMeterNotStoppedError):
        energy_meter.__iter__()


def test_get_sample_on_a_non_stopped_energy_meter_raise_EnergyMeterNotStopdError(energy_meter):
    energy_meter.start()
    with pytest.raises(EnergyMeterNotStoppedError):
        energy_meter.get_sample('toto')


####################
# GET ENERGY TRACE #
####################
def test_iter_on_one_sample_trace_should_return_one_sample(energy_meter):
    energy_meter.start()
    energy_meter.stop()
    samples = []

    for sample in energy_meter:
        samples.append(sample)

    assert len(samples) == 1


@pytest.fixture
def sample1():
    ts = TIMESTAMP_TRACE[0]
    tag = ''
    duration = TIMESTAMP_TRACE[1] - TIMESTAMP_TRACE[0]
    energy = {str(EnergyDomainDevice1Domain1()): DEVICE1_ENERGY_TRACE[1][0] - DEVICE1_ENERGY_TRACE[0][0],
              str(EnergyDomainDevice1Domain2()): DEVICE1_ENERGY_TRACE[1][1] - DEVICE1_ENERGY_TRACE[0][1],
              str(EnergyDomainDevice2Domain1()): DEVICE2_ENERGY_TRACE[1][0] - DEVICE2_ENERGY_TRACE[0][0]}
    return EnergySample(ts, tag, duration, energy)


@pytest.fixture
def sample2():
    ts = TIMESTAMP_TRACE[1]
    tag = ''
    duration = TIMESTAMP_TRACE[2] - TIMESTAMP_TRACE[1]
    energy = {str(EnergyDomainDevice1Domain1()): DEVICE1_ENERGY_TRACE[2][0] - DEVICE1_ENERGY_TRACE[1][0],
              str(EnergyDomainDevice1Domain2()): DEVICE1_ENERGY_TRACE[2][1] - DEVICE1_ENERGY_TRACE[1][1],
              str(EnergyDomainDevice2Domain1()): DEVICE2_ENERGY_TRACE[2][0] - DEVICE2_ENERGY_TRACE[1][0]}
    return EnergySample(ts, tag, duration, energy)


@pytest.fixture
def sample3():
    ts = TIMESTAMP_TRACE[3]
    tag = ''
    duration = TIMESTAMP_TRACE[4] - TIMESTAMP_TRACE[3]
    energy = {str(EnergyDomainDevice1Domain1()): DEVICE1_ENERGY_TRACE[4][0] - DEVICE1_ENERGY_TRACE[3][0],
              str(EnergyDomainDevice1Domain2()): DEVICE1_ENERGY_TRACE[4][1] - DEVICE1_ENERGY_TRACE[3][1],
              str(EnergyDomainDevice2Domain1()): DEVICE2_ENERGY_TRACE[4][0] - DEVICE2_ENERGY_TRACE[3][0]}
    return EnergySample(ts, tag, duration, energy)


@patch('time.time_ns', side_effect=TIMESTAMP_TRACE)
def test_iter_on_one_sample_trace_should_return_correct_values(_mocked_fun, energy_meter, sample1):
    energy_meter.start()
    energy_meter.stop()
    samples = []

    for sample in energy_meter:
        assert_sample_are_equals(sample, sample1)


@patch('time.time_ns', side_effect=TIMESTAMP_TRACE)
def test_iter_on_two_sample_trace_should_return_two_sample(_mocked_fun, energy_meter):
    energy_meter.start()
    energy_meter.record()
    energy_meter.stop()

    samples = []
    for sample in energy_meter:
        samples.append(sample)

    assert len(samples) == 2


@patch('time.time_ns', side_effect=TIMESTAMP_TRACE)
def test_iter_on_two_sample_trace_should_return_correct_values(_mocked_fun, energy_meter, sample1, sample2):
    energy_meter.start()
    energy_meter.record()
    energy_meter.stop()

    for sample, correct_sample in zip(energy_meter, [sample1, sample2]):
        assert_sample_are_equals(sample, correct_sample)


@patch('time.time_ns', side_effect=TIMESTAMP_TRACE)
def test_get_sample_on_a_one_sample_trace_return_correct_values(_mocked_fun, energy_meter, sample1):
    energy_meter.start()
    energy_meter.stop()
    assert_sample_are_equals(energy_meter.get_sample(''), sample1)


@patch('time.time_ns', side_effect=TIMESTAMP_TRACE)
def test_get_sample_on_a_two_sample_trace_with_same_names_return_first_sample(_mocked_fun, energy_meter, sample1):
    energy_meter.start()
    energy_meter.record()
    energy_meter.stop()

    assert_sample_are_equals(energy_meter.get_sample(''), sample1)


@patch('time.time_ns', side_effect=TIMESTAMP_TRACE)
def test_get_sample_by_their_names_return_correct_samples(_mocked_fun, energy_meter, sample1, sample2):
    energy_meter.start('sample1')
    energy_meter.record('sample2')
    energy_meter.stop()

    sample1.tag = 'sample1'
    sample2.tag = 'sample2'

    assert_sample_are_equals(energy_meter.get_sample('sample1'), sample1)
    assert_sample_are_equals(energy_meter.get_sample('sample2'), sample2)


def test_get_sample_that_does_not_exist_raise_SampleNotFoundError(energy_meter):
    energy_meter.start()
    energy_meter.stop()

    with pytest.raises(SampleNotFoundError):
        energy_meter.get_sample('sample1')


@patch('time.time_ns', side_effect=TIMESTAMP_TRACE)
def test_second_start_on_an_energy_meter_should_restart_the_trace(_mocked_fun, energy_meter, sample3):
    energy_meter.start()
    energy_meter.record()
    energy_meter.stop()
    energy_meter.start()
    energy_meter.stop()

    samples = []
    for sample in energy_meter:
        samples.append(sample)

    assert len(samples) == 1
    assert_sample_are_equals(samples[0], sample3)


############
# TEST TAG #
############
def test_define_energy_meter_with_default_tag_create_sample_with_default_tag():

    device1 = MockedEnergyDevice1()
    device1.configure()
    device2 = MockedEnergyDevice2()
    device2.configure()
    meter = EnergyMeter([device1, device2], default_tag='tag')

    meter.start()
    meter.stop()

    sample = meter.get_sample('tag')
    assert sample.tag == 'tag'



# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os.path

from . import EnergyRecorder


class CSVHandler(EnergyRecorder):

    def __init__(self, filename):
        EnergyRecorder.__init__(self)

        self._filename = filename

    def _gen_header(self):
        domain_names = self.trace_buffer[0].energy.keys()
        return 'timestamp;tag;duration;' + ';'.join(domain_names)

    def _gen_sample_line(self, sample):
        domain_names = self.trace_buffer[0].energy.keys()

        line_begining = f'{sample.timestamp};{sample.tag};{sample.duration};'
        energy_values = [str(sample.energy[domain]) for domain in domain_names]
        return line_begining + ';'.join(energy_values)

    def _init_file(self):
        if os.path.exists(self._filename):
            csv_file = open(self._filename, 'a+')
            return csv_file
        else:
            csv_file = open(self._filename, 'w+')
            csv_file.write(self._gen_header() + '\n')
            return csv_file

    def save_data(self):
        """
        Save each trace contained in the buffer and empty the buffer
        """

        csv_file = self._init_file()
        for sample in self.trace_buffer:
            csv_file.write(self._gen_sample_line(sample) + '\n')
        csv_file.close()
        self.trace_buffer = []



 # MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Test to measure energy consumption of a trace only with the energy meter
"""
import pytest

from mock import patch

from pyJoules.energy_device.rapl_device import RaplDevice, RaplPackageDomain, RaplDramDomain
from pyJoules.energy_device.nvidia_device import NvidiaGPUDomain
from pyJoules.energy_meter import EnergyMeter, EnergyContext
from .. utils.rapl_fs import fs_pkg_dram_one_socket
from ..utils.fake_nvidia_api import one_gpu_api
from .. utils.fake_api import CorrectTrace
from ..utils.sample import assert_sample_are_equals

TIMESTAMP_TRACE = [1.1, 2.2, 3.3, 4.4, 5.5]

@patch('pyJoules.energy_handler.EnergyHandler')
@patch('time.time_ns', side_effect=TIMESTAMP_TRACE)
def test_measure_rapl_device_all_domains(mocked_handler, _mocked_time_ns, fs_pkg_dram_one_socket, one_gpu_api):

    domains = [RaplPackageDomain(0), RaplDramDomain(0), NvidiaGPUDomain(0)]

    correct_trace = CorrectTrace(domains, [fs_pkg_dram_one_socket, one_gpu_api], TIMESTAMP_TRACE)  # test

    correct_trace.add_new_sample('foo')  # test
    with EnergyContext(mocked_handler, domains, start_tag='foo') as energy_context:
        correct_trace.add_new_sample('bar')  # test
        energy_context.record(tag='bar')
        correct_trace.add_new_sample('')  # test

    assert mocked_handler.process.call_count == 2   # test

    for correct_sample, processed_arg in zip(correct_trace, mocked_handler.process.call_args_list):  # test
        measured_sample = processed_arg[0][0]  # test

        assert_sample_are_equals(correct_sample, measured_sample)  # test

@patch('pyJoules.energy_handler.EnergyHandler')
@patch('time.time_ns', side_effect=TIMESTAMP_TRACE)
def test_measure_rapl_device_default_values(mocked_handler, _mocked_time_ns, fs_pkg_dram_one_socket, one_gpu_api):

    correct_trace = CorrectTrace([RaplPackageDomain(0), RaplDramDomain(0), NvidiaGPUDomain(0)],
                                 [fs_pkg_dram_one_socket, one_gpu_api], TIMESTAMP_TRACE)  # test

    correct_trace.add_new_sample('start')  # test
    with EnergyContext(mocked_handler) as energy_context:
        correct_trace.add_new_sample('second_tag')  # test
        energy_context.record(tag='second_tag')
        correct_trace.add_new_sample('')  # test

    assert mocked_handler.process.call_count == 2   # test

    for correct_sample, processed_arg in zip(correct_trace, mocked_handler.process.call_args_list):  # test
        measured_sample = processed_arg[0][0]  # test
        assert_sample_are_equals(correct_sample, measured_sample)  # test



# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
import pynvml
from mock import patch
from ....utils.fake_nvidia_api import no_gpu_api, one_gpu_api, two_gpu_api
from pyJoules.energy_device.nvidia_device import NvidiaGPUDevice, NvidiaGPUDomain
from pyJoules.energy_device import NotConfiguredDeviceException
from pyJoules.exception import NoSuchEnergyDeviceError, NoSuchDomainError

##############
# INIT TESTS #
##############
def test_create_NvidiaDevice_with_no_gpu_api_raise_NoSuchEnergyDeviceError(no_gpu_api):
    with pytest.raises(NoSuchEnergyDeviceError):
        NvidiaGPUDevice()

###########################
# AVAILABLE DOMAINS TESTS #
###########################
def test_available_domains_with_no_gpu_api_raise_NoSuchEnergyDeviceError(no_gpu_api):
    with pytest.raises(NoSuchEnergyDeviceError):
        NvidiaGPUDevice.available_domains()


def test_available_domains_with_one_gpu_api_return_correct_values(one_gpu_api):
    returned_values = NvidiaGPUDevice.available_domains()
    correct_values = [NvidiaGPUDomain(0)]
    assert sorted(correct_values) == sorted(returned_values)


def test_available_domains_with_two_gpu_api_return_correct_values(two_gpu_api):
    returned_values = NvidiaGPUDevice.available_domains()
    correct_values = [NvidiaGPUDomain(0), NvidiaGPUDomain(1)]
    assert sorted(correct_values) == sorted(returned_values)


###################
# CONFIGURE TESTS #
###################
def test_configure_device_to_get_second_gpu_energy_with_no_second_gpu_api_raise_NoSuchDomainError(one_gpu_api):
    device = NvidiaGPUDevice()

    with pytest.raises(NoSuchDomainError):
        device.configure([NvidiaGPUDomain(1)])


def test_get_configured_domains_on_non_configured_device_raise_NotConfiguredDeviceException(one_gpu_api):
    device = NvidiaGPUDevice()

    with pytest.raises(NotConfiguredDeviceException):
        device.get_configured_domains()


def test_get_configured_domains_with_default_values_with_one_gpu_api_return_correct_values(one_gpu_api):
    configured_domains = [NvidiaGPUDomain(0)]
    device = NvidiaGPUDevice()
    device.configure()
    assert configured_domains == device.get_configured_domains()


def test_get_configured_domains_with_default_values_on_two_gpu_api_return_correct_values(two_gpu_api):
    configured_domains = [NvidiaGPUDomain(0), NvidiaGPUDomain(1)]
    device = NvidiaGPUDevice()
    device.configure()
    assert configured_domains == device.get_configured_domains()


########################################
# GET ENERGY WITH DEFAULT VALUES TESTS #
########################################
def test_get_default_energy_values_with_one_gpu_api(one_gpu_api):
    device = NvidiaGPUDevice()
    device.configure()
    assert device.get_energy() == [one_gpu_api.domains_current_energy['nvidia_gpu_0']]


def test_get_default_energy_values_with_two_gpu_api(two_gpu_api):
    device = NvidiaGPUDevice()
    device.configure()
    assert device.get_energy() == [two_gpu_api.domains_current_energy['nvidia_gpu_0'],
                                   two_gpu_api.domains_current_energy['nvidia_gpu_1']]


##################################
# CONFIGURE AND GET ENERGY TESTS #
##################################
def test_get_gpu0_energy_with_only_one_gpu_api_return_correct_value(one_gpu_api):
    """
    Create a NvidiaGpuDevice instance on a machine with one gpu
    configure it to monitor the gpu 0
    use the `get_energy` method and check if:
    - the returned list contains one element
    - this element is the energy consumption of the gpu 0
    """
    device = NvidiaGPUDevice()
    device.configure([NvidiaGPUDomain(0)])
    assert device.get_energy() == [one_gpu_api.domains_current_energy['nvidia_gpu_0']]


def test_get_gpu0_energy_with_two_gpu_api_return_correct_value(two_gpu_api):
    """
    Create a NvidiaGpuDevice instance on a machine with two gpu
    configure it to monitor the gpu 0
    use the `get_energy` method and check if:
    - the returned list contains one element
    - this element is the energy consumption of the gpu 0
    """
    device = NvidiaGPUDevice()
    device.configure([NvidiaGPUDomain(0)])
    assert device.get_energy() == [two_gpu_api.domains_current_energy['nvidia_gpu_0']]

def test_get_gpu0_and_gpu1_energy_with_two_gpu_api_return_correct_value(two_gpu_api):
    """
    Create a NvidiaGpuDevice instance on a machine with two gpu
    configure it to monitor the gpu 0 and gpu 1
    use the `get_energy` method and check if:
    - the returned list contains two elements
    - theses elements are the energy consumption of the gpu 0 and 1
    """
    device = NvidiaGPUDevice()
    device.configure([NvidiaGPUDomain(0), NvidiaGPUDomain(1)])
    assert device.get_energy() == [two_gpu_api.domains_current_energy['nvidia_gpu_0'],
                                   two_gpu_api.domains_current_energy['nvidia_gpu_1']]



# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List


from . import EnergyDevice, EnergyDomain
from ..exception import PyJoulesException, NoSuchEnergyDeviceError
import pynvml

class NvidiaGPUDomain(EnergyDomain):

    def __init__(self, device_id):
        EnergyDomain.__init__(self)
        self.device_id = device_id
        self._repr = 'nvidia_gpu_' + str(device_id)

    def __repr__(self):
        return self._repr

    def __eq__(self, other) -> bool:
        return isinstance(other, NvidiaGPUDomain) and self.__repr__() == other.__repr__()

    def __lt__(self, other) -> bool:
        if isinstance(other, NvidiaGPUDomain):
            return self.device_id < other.device_id
        raise ValueError()

    def __gt__(self, other) -> bool:
        if isinstance(other, NvidiaGPUDomain):
            return self.device_id > other.device_id
        raise ValueError()

    def get_device_type(self):
        return NvidiaGPUDevice


class GPUDoesNotSupportEnergyMonitoringError(PyJoulesException):
    """
    Exception raised when a NvidiaDevice is created but the GPU does not support energy monitoring
    """


class NvidiaGPUDevice(EnergyDevice):
    """
    Interface to get energy consumption of GPUs
    """

    def __init__(self):
        """
        :raise NoSuchEnergyDeviceError: if no Nvidia API is available on this machine
        :raise GPUDoesNotSupportEnergyMonitoringError: if the GPU does not support energy monitoring
        """
        EnergyDevice.__init__(self)
        self._handle = None

    def configure(self, domains: List[NvidiaGPUDomain] = None):
        EnergyDevice.configure(self, domains)
        self._handle = [pynvml.nvmlDeviceGetHandleByIndex(domain.device_id) for domain in self._configured_domains]

    def get_energy(self):
        return [pynvml.nvmlDeviceGetTotalEnergyConsumption(handle) for handle in self._handle]

    @staticmethod
    def available_domains() -> List[NvidiaGPUDomain]:
        try:
            pynvml.nvmlInit()
        except pynvml.NVMLError:
            raise NoSuchEnergyDeviceError()

        device_ids = pynvml.nvmlDeviceGetCount()
        domains = map(lambda device_id: NvidiaGPUDomain(device_id), range(device_ids))
        return list(domains)
